--
-- init pgcrypto
--

CREATE EXTENSION pgcrypto;

-- check error handling
select gen_salt('foo');
select digest('foo', 'foo');
select hmac('foo', 'foo', 'foo');
select encrypt('foo', 'foo', 'foo');
