/* contrib/isn/isn--1.0--1.1.sql */

-- complain if script is sourced in psql, rather than via ALTER EXTENSION
\echo Use "ALTER EXTENSION isn UPDATE TO '1.1'" to load this file. \quit

ALTER FUNCTION ean13_in(cstring) PARALLEL SAFE;
ALTER FUNCTION ean13_out(ean13) PARALLEL SAFE;
ALTER FUNCTION isbn13_in(cstring) PARALLEL SAFE;
ALTER FUNCTION ean13_out(isbn13) PARALLEL SAFE;
ALTER FUNCTION ismn13_in(cstring) PARALLEL SAFE;
ALTER FUNCTION ean13_out(ismn13) PARALLEL SAFE;
ALTER FUNCTION issn13_in(cstring) PARALLEL SAFE;
ALTER FUNCTION ean13_out(issn13) PARALLEL SAFE;
ALTER FUNCTION isbn_in(cstring) PARALLEL SAFE;
ALTER FUNCTION isn_out(isbn) PARALLEL SAFE;
ALTER FUNCTION ismn_in(cstring) PARALLEL SAFE;
ALTER FUNCTION isn_out(ismn) PARALLEL SAFE;
ALTER FUNCTION issn_in(cstring) PARALLEL SAFE;
ALTER FUNCTION isn_out(issn) PARALLEL SAFE;
ALTER FUNCTION upc_in(cstring) PARALLEL SAFE;
ALTER FUNCTION isn_out(upc) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION isnlt(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isnle(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isneq(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isnge(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isngt(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isnne(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnle(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isneq(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnge(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isngt(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnne(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isnle(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isneq(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isnge(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isngt(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isnne(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isnle(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isneq(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isnge(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isngt(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isnne(issn, issn) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isnle(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isneq(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isnge(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isngt(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isnne(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION isnlt(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION isnlt(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isnle(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isneq(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isnge(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isngt(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isnne(upc, upc) PARALLEL SAFE;
ALTER FUNCTION isnlt(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isnle(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isneq(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isnge(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isngt(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isnne(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, ean13) PARALLEL SAFE;
ALTER FUNCTION hashean13(ean13) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, isbn13) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, ismn13) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, issn13) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, isbn) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, ismn) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, issn) PARALLEL SAFE;
ALTER FUNCTION btean13cmp(ean13, upc) PARALLEL SAFE;
ALTER FUNCTION btisbn13cmp(isbn13, isbn13) PARALLEL SAFE;
ALTER FUNCTION hashisbn13(isbn13) PARALLEL SAFE;
ALTER FUNCTION btisbn13cmp(isbn13, ean13) PARALLEL SAFE;
ALTER FUNCTION btisbn13cmp(isbn13, isbn) PARALLEL SAFE;
ALTER FUNCTION btisbncmp(isbn, isbn) PARALLEL SAFE;
ALTER FUNCTION hashisbn(isbn) PARALLEL SAFE;
ALTER FUNCTION btisbncmp(isbn, ean13) PARALLEL SAFE;
ALTER FUNCTION btisbncmp(isbn, isbn13) PARALLEL SAFE;
ALTER FUNCTION btismn13cmp(ismn13, ismn13) PARALLEL SAFE;
ALTER FUNCTION hashismn13(ismn13) PARALLEL SAFE;
ALTER FUNCTION btismn13cmp(ismn13, ean13) PARALLEL SAFE;
ALTER FUNCTION btismn13cmp(ismn13, ismn) PARALLEL SAFE;
ALTER FUNCTION btismncmp(ismn, ismn) PARALLEL SAFE;
ALTER FUNCTION hashismn(ismn) PARALLEL SAFE;
ALTER FUNCTION btismncmp(ismn, ean13) PARALLEL SAFE;
ALTER FUNCTION btismncmp(ismn, ismn13) PARALLEL SAFE;
ALTER FUNCTION btissn13cmp(issn13, issn13) PARALLEL SAFE;
ALTER FUNCTION hashissn13(issn13) PARALLEL SAFE;
ALTER FUNCTION btissn13cmp(issn13, ean13) PARALLEL SAFE;
ALTER FUNCTION btissn13cmp(issn13, issn) PARALLEL SAFE;
ALTER FUNCTION btissncmp(issn, issn) PARALLEL SAFE;
ALTER FUNCTION hashissn(issn) PARALLEL SAFE;
ALTER FUNCTION btissncmp(issn, ean13) PARALLEL SAFE;
ALTER FUNCTION btissncmp(issn, issn13) PARALLEL SAFE;
ALTER FUNCTION btupccmp(upc, upc) PARALLEL SAFE;
ALTER FUNCTION hashupc(upc) PARALLEL SAFE;
ALTER FUNCTION btupccmp(upc, ean13) PARALLEL SAFE;
ALTER FUNCTION isbn13(ean13) PARALLEL SAFE;
ALTER FUNCTION ismn13(ean13) PARALLEL SAFE;
ALTER FUNCTION issn13(ean13) PARALLEL SAFE;
ALTER FUNCTION isbn(ean13) PARALLEL SAFE;
ALTER FUNCTION ismn(ean13) PARALLEL SAFE;
ALTER FUNCTION issn(ean13) PARALLEL SAFE;
ALTER FUNCTION upc(ean13) PARALLEL SAFE;
ALTER FUNCTION make_valid(ean13) PARALLEL SAFE;
ALTER FUNCTION make_valid(isbn13) PARALLEL SAFE;
ALTER FUNCTION make_valid(ismn13) PARALLEL SAFE;
ALTER FUNCTION make_valid(issn13) PARALLEL SAFE;
ALTER FUNCTION make_valid(isbn) PARALLEL SAFE;
ALTER FUNCTION make_valid(ismn) PARALLEL SAFE;
ALTER FUNCTION make_valid(issn) PARALLEL SAFE;
ALTER FUNCTION make_valid(upc) PARALLEL SAFE;
ALTER FUNCTION is_valid(ean13) PARALLEL SAFE;
ALTER FUNCTION is_valid(isbn13) PARALLEL SAFE;
ALTER FUNCTION is_valid(ismn13) PARALLEL SAFE;
ALTER FUNCTION is_valid(issn13) PARALLEL SAFE;
ALTER FUNCTION is_valid(isbn) PARALLEL SAFE;
ALTER FUNCTION is_valid(ismn) PARALLEL SAFE;
ALTER FUNCTION is_valid(issn) PARALLEL SAFE;
ALTER FUNCTION is_valid(upc) PARALLEL SAFE;
ALTER FUNCTION isn_weak(boolean) PARALLEL RESTRICTED;
ALTER FUNCTION isn_weak() PARALLEL RESTRICTED;
