;; see also src/tools/editors/emacs.samples for more complete settings

((c-mode . ((c-basic-offset . 4)
            (c-file-style . "bsd")
            (fill-column . 78)
            (indent-tabs-mode . t)
            (tab-width . 4)))
 (nxml-mode . ((fill-column . 78)
               (indent-tabs-mode . nil)))
 (perl-mode . ((perl-indent-level . 4)
               (perl-continued-statement-offset . 2)
               (perl-continued-brace-offset . -2)
               (perl-brace-offset . 0)
               (perl-brace-imaginary-offset . 0)
               (perl-label-offset . -2)
               (indent-tabs-mode . t)
               (tab-width . 4)))
 (sgml-mode . ((fill-column . 78)
               (indent-tabs-mode . nil))))
